﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CapstoneProject.EngineClasses
{
    class Band
    {
        // environment data
        World world;
        protected Cell cell;

        // band data
        int members;

        // stats
        int births = 0;
        int deaths = 0;
        int totalBirths = 0;
        int totalDeaths = 0;

        #region Properties
        public int Members
        {
            get { return members; }
        }
        public int Births
        {
            get { return births; }
        }
        public int Deaths
        {
            get { return deaths; }
        }
        public int TotalDeaths
        {
            get { return totalDeaths; }
        }
        public int TotalBirths
        {
            get { return totalBirths; }
        }
        #endregion

        /// <summary>
        /// 
        /// </summary>
        /// <param name="members"></param>
        /// <param name="world"></param>
        /// <param name="cell"></param>
        public Band(int members, World world, Cell cell)
        {
            this.members = members;
            this.world = world;
            this.cell = cell;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="members"></param>
        /// <param name="world"></param>
        public Band(int members, World world)
        {
            this.members = members;
            this.world = world;
        }

        /// <summary>
        /// This method encapsulates the logic behind the way in which the band selects its cell as well as the functionality of moving bands to the cell that they find optimal
        /// In order to change band behaviour, inherit the band class and overwrite this method with alternate logic
        /// </summary>
        /// <param name="cells"> The world's cell grid that needs to be searched for the best cell</param>
        public void moveToBestCell(ref Cell[,] cells)
        {
            Cell best = cells[0,0];
            // first we must find an empty cell to compare future cells against
            foreach (Cell currCell in cells)
            {
                if (currCell.Occupied == false) // empty cell found => stop searching
                {
                    best = currCell;
                    break;
                }
            }
            foreach (Cell currCell in cells)
            {
                if (currCell.Occupied == false) // if the current cell is free consider it's resources
                {
                    // compare prey score
                    if (currCell.Score[0] > best.Score[0])
                        best = currCell;
                    else if (currCell.Score[0] == best.Score[0]) // prey equal => compare cereal
                    {
                        if (currCell.Score[1] > best.Score[1])
                            best = currCell;
                        else if (currCell.Score[1] == best.Score[1]) // cereal equal => compare farm
                        {
                            if (currCell.Score[2] > best.Score[2])
                                best = currCell;
                            // if both are equal in all ways keep the current best
                        }
                    }
                }
            }
            // once the best cell has been found move band to the cell
            if (best.Occupied == false) // double check that the cell is empty
            {
                best.addBand(this);
                this.cell = best;
            }
        }

        /// <summary>
        /// Calls helper methods in order to determine the energy required and gathered and the cause of these results, deaths or births.
        /// </summary>
        public void runTimeStep(Random birthRandom)
        {
            births = 0;
            deaths = 0;
            // calculate the amount of energy the band was able to obtain for the timestep
            float energyGathered = consume();
            // calculate the births or deaths for the band
            breedOrDie(energyGathered, birthRandom);
            if (members >= 40)
            {
                split();
            }
            if (members == 0) // band has died
            {
                this.cell.removeBand(); // remove link from cell to band
                this.cell = null; // remove link from band to cell
                world.removeBand(this); // tell world that the band is dead
            }
        }

        /// <summary>
        /// Helper method to deal with the bands resource gathering and requirements
        /// </summary>
        /// <returns> Returns the net energy the band has gathered (energy required - energy gathered), negative if they have gathered enough and positive if not</returns>
        protected float consume()
        {
            float energyGathered = 0;
            if (cell.CellBarley.ResourceNAR > cell.CellFarm.ResourceNAR) // only farm if the barley NAR is less than the Farm NAR
            {
                float totalNAR = cell.CellBarley.ResourceNAR + cell.CellPrey.ResourceNAR;
                float timeAvailable = 14 * 60 * 365 * members; // number of minutes available for the band to find food (hours available in a day * number of mins in and hour * days in a year * number of members)
                float energyRequired = members * 2000 * 365; // energy required by band for the year (number of members * requirement for a member per day * days in a year)

                float[] preyResults = new float[] { 0, 0 };
                float preyEnergyRequired = energyRequired * (cell.CellPrey.ResourceNAR / totalNAR); // percentage in diet is the percentage of the NAR value for that resource divided by the total NAR value for all resources
                if (preyEnergyRequired > 0)
                {
                    preyResults = cell.CellPrey.calculateEnergy(preyEnergyRequired, timeAvailable);
                    timeAvailable -= preyResults[1]; // decrease time available by the time spent hunting
                }

                float[] barleyResults = new float[] { 0, 0 };
                float barleyEnergyRequired = energyRequired * (cell.CellBarley.ResourceNAR / totalNAR);
                if (barleyEnergyRequired > 0)
                {
                    barleyResults = cell.CellBarley.calculateEnergy(barleyEnergyRequired, timeAvailable);
                }

                energyGathered = preyResults[0] + barleyResults[0];

                float netEnergy = energyRequired - energyGathered;

                return netEnergy;
            }
            else // transition to farming
            {
                float totalNAR = cell.CellBarley.ResourceNAR + cell.CellPrey.ResourceNAR + cell.CellFarm.ResourceNAR;
                float timeAvailable = 14 * 60 * 365 * members; // number of minutes available for the band to find food (hours available in a day * number of mins in and hour * days in a year * number of members)
                float energyRequired = members * 2000 * 365; // energy required by band for the year (number of members * requirement for a member per day * days in a year)

                float[] preyResults = new float[] { 0, 0 };
                float preyEnergyRequired = energyRequired * (cell.CellPrey.ResourceNAR / totalNAR); // percentage in diet is the percentage of the NAR value for that resource divided by the total NAR value for all resources
                if (preyEnergyRequired > 0)
                {
                    preyResults = cell.CellPrey.calculateEnergy(preyEnergyRequired, timeAvailable);
                    timeAvailable -= preyResults[1]; // decrease time available by the time spent hunting
                }

                float[] barleyResults = new float[] { 0, 0 };
                float barleyEnergyRequired = energyRequired * (cell.CellBarley.ResourceNAR / totalNAR);
                if (barleyEnergyRequired > 0)
                {
                    barleyResults = cell.CellBarley.calculateEnergy(barleyEnergyRequired, timeAvailable);
                    timeAvailable -= barleyResults[1]; // decrease time available by the time spent hunting
                }

                float[] farmResults = new float[] { 0, 0 };
                float farmEnergyRequired = energyRequired * (cell.CellFarm.ResourceNAR / totalNAR);
                if (farmEnergyRequired > 0)
                {
                    farmResults = cell.CellFarm.calculateEnergy(farmEnergyRequired, timeAvailable);
                }

                energyGathered = preyResults[0] + barleyResults[0] + farmResults[0];

                float netEnergy = energyRequired - energyGathered;

                return netEnergy;
            }

            
        }

        /// <summary>
        /// Helper method that determines, based on the bands net energy, the number of member that died or were born for the timestep
        /// </summary>
        /// <param name="netEnergy">The net energy of the band (energy still required)</param>
        protected void breedOrDie(float netEnergy, Random birthRandom)
        {
            if (netEnergy > 1) // band needed more energy than it gathered => deaths
            {
                int currentMembers = members;
                members = Math.Max(0, members - ((int)(netEnergy / (365 * 2000)) + 1));
                deaths = currentMembers - members;
                totalDeaths += deaths;
            }
            else // band gathered enough energy => births
            {
                int currBirths = 0; // total births for this timestep
                for (int i = 0; i < members; i++) // each member has a chance of reproducing
                {
                    if (birthRandom.NextDouble() < 0.02) // 2% chance for a birth. NextDouble returns a value between 0 and 1
                        currBirths++;
                }
                members += currBirths; // increase member count by number of births
                births = currBirths;
                totalBirths += currBirths;
            }
        }

        /// <summary>
        /// Helper method for when the band has reached its splitting capacity. If the world is not full a new band with 20 members is added to the world and the count of this band is decreased by 20
        /// </summary>
        protected void split()
        {
            if (!world.isFull()) // can only split if there is an empty cell
            {
                world.addBand(new Band(20, world));
                members -= 20;
            }
        } 
    }
}
