﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CapstoneProject.EngineClasses
{
    class Cell
    {
        // the x and y coordinate of the cell in the world grid
        int x;
        int y;

        Band band; // the band currently occupying the cell or null if none
        Boolean occupied = false; // the state of the cell, occupied or not

        Habitat cellHabitat; // the habitat that the cell is part of, determines properties of the cell's resources

        // the three resources contained within the cell
        Prey prey;
        Barley barley;
        Farming farm;

        protected float[] score; // float array containing the cell's score for each food resource [prey, barley, farm]

        #region Properties
        public float[] Score
        {
            get { return score; }
        }

        public int X
        {
            get { return x; }
        }

        public int Y
        {
            get { return y; }
        }

        public Band CellBand
        {
            get { return band; }
        }

        public Boolean Occupied
        {
            get { return occupied; }
        }

        public Prey CellPrey
        {
            get { return prey; }
        }

        public Barley CellBarley
        {
            get { return barley; }
        }

        public Farming CellFarm
        {
            get { return farm; }
        }

        public Habitat CellHabitat
        {
            get { return cellHabitat; }
            set { cellHabitat = value; }
        }
        #endregion

        /// <summary>
        /// The only possible constructor for the cell, each cell must have an x and y within the world and a habitat it belongs to
        /// </summary>
        /// <param name="x"> The cell's x coordinate in the world grid</param>
        /// <param name="y"> The cell's y coordinate in the world grid</param>
        /// <param name="habitat"> The habitat that the cell belongs to, determines properties of the cell's resources</param>
        public Cell(int x, int y, Habitat habitat)
        {
            this.x = x;
            this.y = y;
            this.cellHabitat = habitat;
            // initialize each of the cell's resources using the habitat
            prey = new Prey(cellHabitat);
            barley = new Barley(cellHabitat);
            farm = new Farming(cellHabitat);

            score = new float[3]; // intialize score array
        }

        /// <summary>
        /// Method to update the cell's score array, each of the cell's food resources have their score calculated and then the score array is updated for each resource
        /// </summary>
        public void calculateScore()
        {
            // update NAR values for all resources
            prey.calculateNAR();
            barley.calculateNAR();
            farm.calculateNAR();
            // set appropriate score values
            score[0] = prey.ResourceNAR;
            score[1] = barley.ResourceNAR;
            score[2] = farm.ResourceNAR;
        }

        /// <summary>
        /// Sets occupied to true and sets the band reference of this cell to the input band. Note this method does not alter the input band's cell reference
        /// </summary>
        /// <param name="newBand"> The band that this cell should consider itself occupied by</param>
        public void addBand(Band newBand)
        {
            occupied = true;
            band = newBand;
        }

        /// <summary>
        /// Sets the cells current band to null if it wasn't already, and sets occupied to false. Note that the reference in the band class is not set to null and so could still point to this cell
        /// </summary>
        public void removeBand()
        {
            if (band != null)
            {
                occupied = false;
                band = null;
            }
        }

        /// <summary>
        /// Calls the grow method on each of the cell's food resources
        /// </summary>
        public void grow()
        {
            prey.grow();
            barley.grow();
            farm.grow();
        }
    }
}
