﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace CapstoneProject.EngineClasses
{
    // This class is responsible for the reading and writing of simulation data to text files. This data is used in graph construction and can be read by users of the simulation
    class FileManager
    {
        // File name of the output files
        const string preyFile = "CellPrey.txt";
        const string barleyFile = "CellBarley.txt";
        const string farmFile = "CellFarm.txt";
        const string statsFile = "StatsFile.txt";
        const string cellFile = "CellFile.txt";

        StreamWriter preyFileWriter; // contains prey data for each cell
        StreamWriter barleyFileWriter; // contains barley data for each cell
        StreamWriter farmFileWriter; // contains farm data for each cell
        StreamWriter statsFileWriter; // contains the summarized data for the world
        StreamWriter cellFileWriter; // contains all data for each cell

        // stat values
        float averagePreyAll = 0;
        float averageBarleyAll = 0;
        float averageFarmAll = 0;
        float totalDeaths = 0;
        float totalBirths = 0;
        float totalPopulation = 0;
        float birthsByTurn = 0;
        float deathsByTurn = 0;

        public FileManager()
        {
            // first clear all the text files
            clearFiles();
            // then create the StreamWriters for each file and set them to append
            preyFileWriter = new StreamWriter(preyFile, true); // true parameter sets writer to append mode
            barleyFileWriter = new StreamWriter(barleyFile, true);
            farmFileWriter = new StreamWriter(farmFile, true);
            statsFileWriter = new StreamWriter(statsFile, true);
            cellFileWriter = new StreamWriter(cellFile, true);

        }

        /// <summary>
        /// Helper function which overwrites the data in all the output files with an empty character in order to reset the text files
        /// </summary>
        private void clearFiles()
        {
            using (StreamWriter file = new StreamWriter(preyFile, false)) // false parameter sets streamwriter to overwrite
            {
                file.WriteLine("");
            }
            using (StreamWriter file = new StreamWriter(barleyFile, false)) // using causes the streamwriter to be disposed immediately after leaving scope
            {
                file.WriteLine("");
            }
            using (StreamWriter file = new StreamWriter(farmFile, false))
            {
                file.WriteLine("");
            }
            using (StreamWriter file = new StreamWriter(statsFile, false))
            {
                file.WriteLine("");
            }
            using (StreamWriter file = new StreamWriter(cellFile, false))
            {
                file.WriteLine("");
            }
        }

        /// <summary>
        /// The function for writing all of the worlds data and statistics to the data logs
        /// </summary>
        /// <param name="cells">The cell grid of the world</param>
        /// <param name="bands">the list of bands in the world</param>
        /// <param name="timestep">the current timestep (passed in by the engine)</param>
        /// <param name="worldsTotalBirths">the total births that occurred in now-dead bands</param>
        /// <param name="worldsTotalDeaths">the total deaths that occurred in now-dead bands</param>
        public void WriteWorldToFiles(Cell[,] cells, List<Band> bands, int timestep, int worldsTotalBirths, int worldsTotalDeaths)
        {
            string preyString;
            string barleyString;
            string farmString;
            string statString;
            string cellString;

            float totalPrey = 0;
            float totalBarley = 0;
            float totalFarm = 0;
            // write the cell statistics for each cell
            foreach (Cell cell in cells)
            {
                preyString = "[ TS:" + timestep + " COORD: " + cell.X + "," + cell.Y + " Val: " + cell.CellPrey.CurrentValue + " NAR: " + cell.CellPrey.ResourceNAR + " ]";
                barleyString = "[ TS:" + timestep + " COORD: " + cell.X + "," + cell.Y + " Val: " + cell.CellBarley.CurrentValue + " NAR: " + cell.CellBarley.ResourceNAR + " ]";
                farmString = "[ TS:" + timestep + " COORD: " + cell.X + "," + cell.Y + " Val: " + cell.CellFarm.CurrentValue + " NAR: " + cell.CellFarm.ResourceNAR + " ]";
                cellString = "[ TS:" + timestep + " COORD: " + cell.X + "," + cell.Y + " PreyVal: " + cell.CellPrey.CurrentValue + " PreyNAR: " + cell.CellPrey.ResourceNAR
                    + " BarleyVal: " + cell.CellBarley.CurrentValue + " BarleyNAR: " + cell.CellBarley.ResourceNAR + " FarmVal: " + cell.CellFarm.CurrentValue + " FarmNAR: " +  cell.CellFarm.ResourceNAR + " ]";
                // add the cell's value to the cumulative values
                totalPrey += cell.CellPrey.CurrentValue;
                totalBarley += cell.CellBarley.CurrentValue;
                totalFarm += cell.CellFarm.CurrentValue;

                preyFileWriter.WriteLine(preyString);
                barleyFileWriter.WriteLine(barleyString);
                farmFileWriter.WriteLine(farmString);
                cellFileWriter.WriteLine(cellString);
            }
            // reset the non-cumulative values
            deathsByTurn = 0;
            birthsByTurn = 0;
            totalPopulation = 0;
            totalBirths = 0;
            totalDeaths = 0;
            // get stats from each of the bands currently alive
            foreach (Band band in bands)
            {
                totalPopulation += band.Members;
                totalBirths += band.TotalBirths;
                totalDeaths += band.TotalDeaths;
                birthsByTurn += band.Births;
                deathsByTurn += band.Deaths;
            }
            // add stats from the bands that have died and passed their stat data to the world object
            totalBirths += worldsTotalBirths;
            totalDeaths += worldsTotalDeaths;
            // calculate average
            averagePreyAll = totalPrey / cells.Length; // amount/ length of the cell array (total number of cells)
            averageBarleyAll = totalBarley / cells.Length;
            averageFarmAll = totalFarm / cells.Length;

            // statistic file output format: [ TS: (timestep) ATP: (Avg total prey) ATB: (Avg total Barley) ATF: (Avg total farm) TP: (total population) TB: (total births) TD: (total deaths) BBT: (births by turn) DBT (deaths by turn)]
            statString = "[ TS: " + timestep + " ATP: " + averagePreyAll + " ATB: " + averageBarleyAll + " ATF: " + averageFarmAll + " TP: " + totalPopulation + " TB: " + totalBirths + " TD: " + totalDeaths + 
                " BBT: " + birthsByTurn + " DBT: " + deathsByTurn + " ]" ;

            statsFileWriter.WriteLine(statString);
            // make sure the writers are empty and all the data for the timestep has been written to the files
            preyFileWriter.Flush();
            barleyFileWriter.Flush();
            farmFileWriter.Flush();
            statsFileWriter.Flush();
            cellFileWriter.Flush();
        }

        /// <summary>
        /// This is the method used to gather the data we have written to the files and place it in the lists that are passed in. These lists are passed from the statistics screen and are used for drawing of the graphs
        /// </summary>
        /// <param name="ATPValues">A reference to be updated with the Average total Prey</param>
        /// <param name="ATBValues">A reference to be updated with the Average total Barley</param>
        /// <param name="ATFValues">A reference to be updated with the Average total farm</param>
        /// <param name="TPValues">A reference to be updated with the total population values</param>
        /// <param name="TBValues">A reference to be updated with the total births</param>
        /// <param name="TDValues">A reference to be updated with the total deaths</param>
        /// <param name="BBTValues">A reference to be updated with the births by turn</param>
        /// <param name="DBTValues">A reference to be updated with the deaths by turn</param>
        public void ReadFromFile(ref List<float> ATPValues, ref List<float> ATBValues, ref List<float> ATFValues, ref List<float> TPValues, ref List<float> TBValues, ref List<float> TDValues, ref List<float> BBTValues, ref List<float> DBTValues)
        {
            closeWriterStreams(); // cannot access files if our writers are accessing them
            string[] lines = System.IO.File.ReadAllLines(statsFile);
            for (int i = 0; i < lines.Length; i++)
            {
                string[] line;
                line = lines[i].Split(' ');
                if (line.Length > 18) // ensure the line contains a proper stat output and isn't empty
                {
                    try
                    {
                        ATPValues.Add((float)Convert.ToDouble(line[4].Trim()));
                        ATBValues.Add((float)Convert.ToDouble(line[6].Trim()));
                        ATFValues.Add((float)Convert.ToDouble(line[8].Trim()));
                        TPValues.Add((float)Convert.ToDouble(line[10].Trim()));
                        TBValues.Add((float)Convert.ToDouble(line[12].Trim()));
                        TDValues.Add((float)Convert.ToDouble(line[14].Trim()));
                        BBTValues.Add((float)Convert.ToDouble(line[16].Trim()));
                        DBTValues.Add((float)Convert.ToDouble(line[18].Trim()));
                    }
                    catch (FormatException e)
                    {
                        Console.Error.WriteLine(e);
                        Console.WriteLine("Error reading in stats");
                    }
                }
            }
            reopenWriterStreams(); // reopen the writer streams so that when we return to the simulation they can continue recording data
        }

        // used for closing the filewriter streams, used for creating a new FileManager of when we need to read from the files instead of write
        public void closeWriterStreams()
        {
            // ensure writers are empty
            preyFileWriter.Flush();
            barleyFileWriter.Flush();
            farmFileWriter.Flush();
            cellFileWriter.Flush();
            statsFileWriter.Flush();
            // dispose writers
            preyFileWriter.Dispose();
            barleyFileWriter.Dispose();
            farmFileWriter.Dispose();
            cellFileWriter.Dispose();
            statsFileWriter.Dispose();
        }

        // used once we have closed our streams, this allows us to reopen the streams in the same manner they are opened in the constructor
        private void reopenWriterStreams()
        {
            // create the StreamWriters for each file and set them to append
            preyFileWriter = new StreamWriter(preyFile, true); // true parameter sets writer to append mode
            barleyFileWriter = new StreamWriter(barleyFile, true);
            farmFileWriter = new StreamWriter(farmFile, true);
            statsFileWriter = new StreamWriter(statsFile, true);
            cellFileWriter = new StreamWriter(cellFile, true);
        }
    }
}
