﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CapstoneProject.EngineClasses
{
    abstract class FoodResource
    {
        protected float searchTime; // time taken to find
        protected float handleTime; // time taken to gather/catch

        protected float searchCost = 4; // energy cost to find
        protected float handleCost = 6; // energy cost to gather/catch

        protected float energy; // energy amount provided by resource
        protected float maxValue; // maximum amount that the cell can sustain
        protected float growthRate; // rate at which the resource grows in the given cell
        protected float viableValue; // the minimum amount of resource required in order for the cell to attempt repopulating other cells (also the initial value for repopulated cells)

        protected float currentValue; // current value of resource measured in number(prey) or kilos (cereal) /km^2
        protected float NAR; // Net acquisition rate. Total amount of energy gained from resource minus energy used acquiring it

        #region Properties
        public float ViableValue
        {
            get { return viableValue; }
        }
        public float ResourceNAR
        {
            get { return NAR; }
        }
        public float CurrentValue
        {
            get { return currentValue; }
        }
        #endregion

        /// <summary>
        /// 
        /// </summary>
        /// <param name="energy"></param>
        /// <param name="handleTime"></param>
        /// <param name="maxValue"></param>
        /// <param name="growthRate"></param>
        /// <param name="viableValue"></param>
        public FoodResource(float energy, float handleTime, float maxValue, float growthRate, float viableValue)
        {
            this.energy = energy;
            this.handleTime = handleTime;
            this.maxValue = maxValue;
            this.growthRate = Convert.ToInt64(Math.Pow(Math.E, growthRate));
            this.viableValue = viableValue;

            currentValue = maxValue;
            //currentValue = maxValue/2; // growth testing - un-comment to set the starting values to half the cell max
            /*if (maxValue == 1620) // repopulation testing - un-comment to set medium zones starting values to 0 and watch as they repopulate
                currentValue = 0;*/
        }

        /// <summary>
        /// calculates the new current value for the resource using the equation provided (p(t+1) = p(t) * (maxVal*e^growth)/(maxVal*(1-searchTime(1-e^growth)))
        /// </summary>
        public void grow()
        {
            float nextValue;
            nextValue = currentValue * (maxValue * growthRate)/ (maxValue - (currentValue *(1-growthRate)));
            if (nextValue < maxValue) // ensure the value cannot go beyond the maximum amount sustainable by the cell
                currentValue = nextValue;
            else
                currentValue = maxValue;
        }

        /// <summary>
        /// Helper function to calculate the search time for the current timestep
        /// </summary>
        private void calculateSearchTime()
        {
            float currSearchTime = 1/(0.5f * 0.0175f * 2 * (currentValue/300));
            if (float.IsInfinity(currSearchTime)) // prevent the extinction of a resource from returning infinite search times
            {
                currSearchTime = 100000; // instead the search time is made very high
            }
            searchTime = currSearchTime;
        }

        /// <summary>
        /// calculate the current NAR for the timestep
        /// </summary>
        public void calculateNAR()
        {
            calculateSearchTime(); // NAR equation requires the search time for the resource
            float currNAR = (energy / (searchTime + handleTime)) - ((searchTime/(searchTime + handleTime)) * searchCost) - ((handleTime/(searchTime + handleTime)) * handleCost);
            //Console.WriteLine(((searchTime / (searchTime + handleTime)) * searchCost));
            NAR = currNAR;
        }

        /// <summary>
        /// Calculates the amount of energy the band is able to obtain from this resource
        /// </summary>
        /// <param name="energyNeeded">The energy the band requires from this resource</param>
        /// <param name="timeAvail">Time that the band has left for gathering resources</param>
        /// <returns>Returns a float array of length 2, the first value is the energy gathered from the resource and the second is the time spent gathering the resource</returns>
        public float[] calculateEnergy(float energyNeeded, float timeAvail)
        {
            float[] output = new float[] {0,0};
            if (currentValue == 0)
                return output;
            float catchable = Math.Min(currentValue, timeAvail / (searchTime + handleTime)); // cant catch more than available or more than time allows
            float needed = (energyNeeded / NAR) / handleTime;
            float caught = Math.Min(catchable, needed);
            //Console.WriteLine("needed: " + needed + " caught: " + caught + " catchable: " + catchable + "");

            output[0] = NAR * (searchTime + handleTime) * caught;
            output[1] = (searchTime + handleTime) * caught;
            currentValue = Math.Max(0, currentValue - caught);

            return output;
        }

        /// <summary>
        /// Function that tries to repopulate a resource when it has been depleted
        /// </summary>
        /// <param name="viableCells"> the number of cells adjacent that contain a viable amount of the resource</param>
        /// <param name="chance"> the random number (between 1 and 10) that must be beaten in order for repopulation to occur</param>
        public void repopulate(int viableCells, int chance)
        {
            if (chance < viableCells) // 10% chance for each neighbouring cell to repopulate this one
            {
                // repopulation succeeds
                currentValue = viableValue-1; // cannot repopulate other cells if it has been repopulated this turn so value set to just under viable value
            }
        }
    }
}
