﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CapstoneProject.EngineClasses
{
    class World
    {
        #region Attributes
        int length, height; // length and height of the cell grid that makes up the world

        Cell[,] cells; // the array of cells

        Habitat[] habitats; // array of created habitats to be assigned to cells
        List<Band> bands; // all currently alive bands
        List<Band> newBands; // bands that have been created this timestep
        List<Band> deadBands; // bands that have died this timestep

        // stats - the world records the stats of bands that have died out in order to maintain the statistic data that would be lost once the band is removed
        int totalDeaths = 0;
        int totalBirths = 0;

        int initialBands = 3; // the initial number of bands placed into the world
        int initialBandSize = 20; // the number of members that the bands created at world start begin with
        #endregion

        #region Properties
        public Habitat[] Habitats
        {
            get { return habitats; }
        }

        public Cell[,] Cells
        {
            get { return cells; }
        }
        public List<Band> Bands
        {
            get { return bands; }
        }
        public int TotalDeaths
        {
            get { return totalDeaths; }
        }
        public int TotalBirths
        {
            get { return totalBirths; }
        }
        #endregion

        // Constructors
        /// <summary>
        /// Default constructor. Creates a 9x9 world with 3 bands of size 20, and the 3 default habitats
        /// </summary>
        public World()
        {
            // Create the habitats
            Habitat lush = new Habitat(this, "Lush", 2400,180000,0);
            Habitat med = new Habitat(this, "Medium", 1620, 120000, 0);
            Habitat desert = new Habitat(this, "Desert", 780, 60000, 0);
            // Add habitats to the habitat array
            habitats = new Habitat[3]{ lush, med, desert};

            // set height and length values
            length = 9;
            height = 9;

            // create grid
            cells = new Cell[length, height];
            createCells();

            // create and place initial bands
            bands = new List<Band>();
            placeInitialBands();

            // create the lists to hold certain bands during timesteps
            newBands = new List<Band>();
            deadBands = new List<Band>();
        }

        /// <summary>
        /// Custom world size contructor. Takes in two parameters to create a NxM world with 3 bands of size 20 and the 3 default habitats
        /// </summary>
        /// <param name="worldLength">the length of the world's grid</param>
        /// <param name="worldHeight">the height of the world's grid</param>
        public World(int worldLength, int worldHeight)
        {
            // Create the habitats
            Habitat lush = new Habitat(this, "Lush", 2400, 180000, 0);
            Habitat med = new Habitat(this, "Medium", 1620, 120000, 0);
            Habitat desert = new Habitat(this, "Desert", 780, 60000, 0);
            // Add habitats to the habitat array
            habitats = new Habitat[3] { lush, med, desert };

            // set height and length values
            length = worldLength;
            height = worldHeight;

            // create grid
            cells = new Cell[length, height];
            createCells();

            // create and place initial bands
            bands = new List<Band>();
            placeInitialBands();

            // create the lists to hold certain bands during timesteps
            newBands = new List<Band>();
            deadBands = new List<Band>();
        }

        /// <summary>
        /// Constructs the world using a custom size as well as a custom number and size of bands
        /// </summary>
        /// <param name="worldLength">The length of the world's grid</param>
        /// <param name="worldHeight">The height of the world's grid</param>
        /// <param name="initialBands">The custom number of initial bands placed in the world</param>
        /// <param name="startingMembers">The custom number of members that each initial band starts with</param>
        public World(int worldLength, int worldHeight, int initialBands, int startingMembers)
        {
            // Create the habitats
            Habitat lush = new Habitat(this, "Lush", 2400, 180000, 0);
            Habitat med = new Habitat(this, "Medium", 1620, 120000, 0);
            Habitat desert = new Habitat(this, "Desert", 780, 60000, 0);
            // Add habitats to the habitat array
            habitats = new Habitat[3] { lush, med, desert };

            // set height and length values
            length = worldLength;
            height = worldHeight;

            // create grid
            cells = new Cell[length, height];
            createCells();

            this.initialBands = initialBands;
            initialBandSize = startingMembers;

            // create and place initial bands
            bands = new List<Band>();
            placeInitialBands();

            // create the lists to hold certain bands during timesteps
            newBands = new List<Band>();
            deadBands = new List<Band>();
        }

        // Constructor Helpers
        /// <summary>
        /// A helper method for creating all the cells in the world grid. Only handles 3 habitats, and divides the world evenly along its length giving float values to the earlier zones
        /// </summary>
        private void createCells()
        {
            // divide world into three sections defined by the 2 boundaries that seperate each section
            int firstBoundary = Convert.ToInt32((Math.Ceiling((decimal)length/3)));
            int rem = length - firstBoundary;
            //int secondBoundary = Convert.ToInt32((Math.Floor((decimal)length / 3)))*2;
            int secondBoundary = firstBoundary + Convert.ToInt32((Math.Ceiling((decimal)rem / 2)));
            for (int i = 0; i < length; i++)
            {
                for (int k = 0; k < height; k++)
                {
                    // set appropriate habitat based on i (x value of the cell)
                    if (i < firstBoundary)
                        cells[i, k] = new Cell(i, k, habitats[0]);
                    else if (i >= firstBoundary && i < secondBoundary)
                        cells[i, k] = new Cell(i, k, habitats[1]);
                    else
                        cells[i, k] = new Cell(i, k, habitats[2]);
                }
            }
        }

        /// <summary>
        /// Helper method for placing the intial bands within the world
        /// </summary>
        private void placeInitialBands()
        {
            for (int i = 0; i < initialBands; i++)
            {
                Random rand = new Random();
                // create a random x and y coordinate
                int row = rand.Next(0,length);
                int col = rand.Next(0,height);
                // attempt to place band
                if (cells[row, col].Occupied == false) // make sure cell is empty
                {
                    Band toAdd = new Band(initialBandSize, this, cells[row, col]);
                    cells[row, col].addBand(toAdd);
                    bands.Add(toAdd);
                }
                    // cell is occupied, band must still be placed somewhere so decrement i
                else
                    i--;

            }
        }
        // Mutators
        /// <summary>
        /// public method for adding a band to the new bands list so that it is appended to the bands list and given a cell
        /// </summary>
        /// <param name="band">Band that needs to be added to the alive bands list</param>
        public void addBand(Band band)
        {
            newBands.Add(band);
        }

        /// <summary>
        /// public method for adding a band to the dead bands list so that the band is removed during the timestep
        /// </summary>
        /// <param name="band">The band that has died and needs to be removed from alive bands list</param>
        public void removeBand(Band band)
        {
            deadBands.Add(band);
            // record the bands stats before it is removed
            totalBirths += band.TotalBirths;
            totalDeaths += band.TotalDeaths;
        }

        /// <summary>
        /// 
        /// </summary>
        public void runTimeStep()
        {
            Random rand = new Random();
            // attempt repopulation
            foreach (Cell cell in cells)
            {
                if (cell.CellPrey.CurrentValue == 0) // check whether prey requires repopulation
                {
                    repopulateCellPrey(cell, rand.Next(0, 10));
                    
                }
                if (cell.CellBarley.CurrentValue == 0) // check whether barley requires repopulation
                {
                    repopulateCellBarley(cell, rand.Next(0, 10));
                }
            }
            // update all the cells
            foreach (Cell cell in cells)
            {
                cell.grow(); // grow each resource in the cell
                cell.calculateScore(); // calculate NAR values for each resource in cell
                cell.removeBand(); // all bands are removed from their cells and the cells are then allocated to the oldest bands (earliest in alive bands list) 
            }
            Random bandBirthRandom = new Random(); // random used to give each band a random number for its breedOrDie function
            // place bands in the optimal cells and run timesteps
            foreach (Band band in bands)
            {
                band.moveToBestCell(ref cells); // alive bands move to the best cell for the next timestep
                band.runTimeStep(bandBirthRandom); // then they run their timestep calculations for timestep based on their new cell
            }
            // new bands are then processed
            foreach (Band newband in newBands)
            {
                bands.Add(newband); // new bands are added to the end of the bands list
                newband.moveToBestCell(ref cells); // then they find their new home
                // newly formed bands don't perform their timestep functions as the are created at the end of the timestep period
            }
            newBands.Clear(); // new band list is cleared once they have all been added for this timestep
            // finally dead bands are processed
            foreach (Band deadband in deadBands)
            {
                bands.Remove(deadband); // deadbands are removed from the alive bands list
            }
            deadBands.Clear(); // deadband list is cleared once all deadbands have been processed for this timestep
        }

        /// <summary>
        /// Helper method for determining the number of cells surrounding the input cell that have a chance of repopulating the input cell's Prey
        /// </summary>
        /// <param name="cell">The cell that needs to be repopulated</param>
        /// <param name="chance">A random integer value that must be beat in order for the cell to be repopulated</param>
        private void repopulateCellPrey(Cell cell, int chance)
        {
            int xMin, xMax;
            int yMin, yMax;
            int viableCount = 0; // the number of adjacent cells with a viable population
            // X - determine the xmin and xmax values of neighbouring cells
            if (cell.X == 0)
            {
                xMin = cell.X;
                xMax = cell.X + 1;
            }
            else if (cell.X == length-1)
            {
                xMin = cell.X - 1;
                xMax = cell.X;
            }
            else
            {
                xMin = cell.X - 1;
                xMax = cell.X + 1;
            }
            // Y - determine the ymin and ymax values of neighbouring cells
            if (cell.Y == 0)
            {
                yMin = cell.Y;
                yMax = cell.Y + 1;
            }
            else if (cell.Y == height-1)
            {
                yMin = cell.Y - 1;
                yMax = cell.Y;
            }
            else
            {
                yMin = cell.Y - 1;
                yMax = cell.Y + 1;
            }
            // check each of the cells surrounding the input cell for viable population and if they have a viable population increment the viableCount
            for (int x = xMin; x <= xMax; x++)
            {
                for (int y = yMin; y <= yMax; y++)
                {
                    if (cells[x, y].CellPrey.CurrentValue >= cells[x, y].CellPrey.ViableValue)
                        viableCount++;
                }
            }
            // attempt repopulation of cell using the gathered data
            cell.CellPrey.repopulate(viableCount, chance);
        }

        /// <summary>
        /// Helper method for determining the number of cells surrounding the input cell that have a chance of repopulating the input cell's Barley
        /// </summary>
        /// <param name="cell">The cell that needs to be repopulated</param>
        /// <param name="chance">A random integer value that must be beat in order for the cell to be repopulated</param>
        private void repopulateCellBarley(Cell cell, int chance)
        {
            int xMin, xMax;
            int yMin, yMax; // the number of adjacent cells with a viable population
            int viableCount = 0;
            // X - determine the xmin and xmax values of neighbouring cells
            if (cell.X == 0)
            {
                xMin = cell.X;
                xMax = cell.X + 1;
            }
            else if (cell.X == length-1)
            {
                xMin = cell.X - 1;
                xMax = cell.X;
            }
            else
            {
                xMin = cell.X - 1;
                xMax = cell.X + 1;
            }
            // Y - determine the ymin and ymax values of neighbouring cells
            if (cell.Y == 0)
            {
                yMin = cell.Y;
                yMax = cell.Y + 1;
            }
            else if (cell.Y == height-1)
            {
                yMin = cell.Y - 1;
                yMax = cell.Y;
            }
            else
            {
                yMin = cell.Y - 1;
                yMax = cell.Y + 1;
            }
            // check each of the cells surrounding the input cell for viable population and if they have a viable population increment the viableCount
            for (int x = xMin; x <= xMax; x++)
            {
                for (int y = yMin; y <= yMax; y++)
                {
                    if (cells[x, y].CellBarley.CurrentValue >= cells[x, y].CellBarley.ViableValue)
                        viableCount++;
                }
            }
            // attempt repopulation of cell using the gathered data
            cell.CellBarley.repopulate(viableCount, chance);
        }
        // Observers
        /// <summary>
        /// Returns true if every cell is occupied otherwise false
        /// </summary>
        /// <returns></returns>
        public Boolean isFull()
        {
            if ((newBands.Count + bands.Count) >= length * height)
                return true;
            else
                return false;
        }

    }
}
