﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using CapstoneProject.EngineClasses;

namespace CapstoneProject.GUI
{
    public partial class GridGUIForm : Form
    {
        Engine engine; // contains the data from the simulation
        Grid grid; // contains the methods for the visualisation of the simulation
        FormContainer container;

        Boolean showingNAR = false; // boolean related to whether we are currently showing the NAR value or the total amount value

        // variables for automatic running
        Boolean runAutomatically = false; // state - whether we are running automatically or not
        private object autolock = new object(); // lock - to prevent errors occuring by calling methods while a timestep is being calculated
        int speed = 1; // current speed at which we are automatically running timesteps

        public GridGUIForm(FormContainer container)
        {
            InitializeComponent();

            this.container = container; // set the container form that contains this form

            engine = new Engine(); // create a new engine - creates a new default world
            grid = new Grid(this, 64, 64, panel1.Right + 50, 25, 9, 9); // create a new grid - default world size
            displayWorldData(); // display the grid
        }

        private void Form1_Load(object sender, EventArgs e)
        {

        }

        /// <summary>
        ///  update the visual representation of the simulation
        /// </summary>
        public void displayWorldData()
        {
            grid.updateGrid(engine.returnCellGrid(), showingNAR); // tell the grid buttons to update themselves using the worlds data
        }

        /// <summary>
        ///  method used by the container form to pass the statistics data to the statistics form
        /// </summary>
        /// <param name="timestep">A reference to be updated with the current timestep</param>
        /// <param name="ATPValues">A reference to be updated with the Average total Prey</param>
        /// <param name="ATBValues">A reference to be updated with the Average total Barley</param>
        /// <param name="ATFValues">A reference to be updated with the Average total farm</param>
        /// <param name="TPValues">A reference to be updated with the total population values</param>
        /// <param name="TBValues">A reference to be updated with the total births</param>
        /// <param name="TDValues">A reference to be updated with the total deaths</param>
        /// <param name="BBTValues">A reference to be updated with the births by turn</param>
        /// <param name="DBTValues">A reference to be updated with the deaths by turn</param>
        public void worldData(ref int timestep, ref List<float> ATPValues, ref List<float> ATBValues, ref List<float> ATFValues, ref List<float> TPValues, ref List<float> TBValues, ref List<float> TDValues, ref List<float> BBTValues, ref List<float> DBTValues)
        {
            engine.worldData(ref timestep, ref ATPValues, ref ATBValues, ref ATFValues, ref TPValues, ref TBValues, ref TDValues, ref BBTValues, ref DBTValues); // call the method in the engine
        }

        private void ManualTimestepButton_Click(object sender, EventArgs e)
        {
            lock (autolock) // make sure the simulation is in a safe state and not running automatically
            {
                runAutomatically = false;
            }
            string text = timeStepTextBox.Text; // get the value set in the timestep size textbox
            int timestep;
            try // try to convert the text to an integer
            {
                timestep = Convert.ToInt32(text);
                timeStepTextBox.BackColor = Color.White;
            }
            catch (FormatException exc) // catch any exceptions
            {
                if (text == "") // no text was entered, timestep default of 1
                    timestep = 1;
                else // the text was not an integer
                {
                    timeStepTextBox.BackColor = Color.Red;
                    Console.Error.WriteLine(exc);
                    MessageBox.Show("The specified timestep is not in the correct format");
                    timestep = 0;
                }
            }
            if (timestep < 0) timestep = 0; // prevent negative input
            if (timestep > 100) timestep = 100; // prevent too many steps in a single timestep
            for (int i = 0; i < timestep; i++) // run timesteps for the specified number
            {
                engine.runTimeStep(); 
            }
            displayWorldData(); // update visual representation once all timesteps are run
        }

        // action taken when exit menustrip button is clicked - exit program
        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        // action taken when show statistics screen menustrip button is clicked - call appropriate container form method
        private void showStatsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            pauseButton_Click(pauseButton, null);
            container.showStatsScreen();
        }

        // action taken when the run automatically button is clicked
        private void runAutoButton_Click(object sender, EventArgs e)
        {
            lock (autolock) // obtain lock
            {
                runAutomatically = true; // set state
                // disable buttons
                ManualTimestepButton.Enabled = false;
                trackBar1.Enabled = false;
                runAutoButton.Enabled = false;
                timeStepTextBox.Enabled = false;
                // show buttons
                pauseButton.Visible = true;
            }
            backgroundWorker2.RunWorkerAsync(); // start the background worker (thread)
        }

        // method for running a timestep and then updating the visual representation - used by the background worker
        private void updateGrid()
        {
            engine.runTimeStep();
            displayWorldData();
        }

        // the work the thread does is sleep as it can't actually edit the grid while working
        private void backgroundWorker2_DoWork(object sender, DoWorkEventArgs e)
        {
            System.Threading.Thread.Sleep(1000/speed); // sleep duration is determined by the speed variable. 1 second at speed setting 1
        }

        // once it is done working (sleeping) this method is called, from this method we are able to edit the grid and so we do
        private void backgroundWorker2_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            lock (autolock)
            {
                if (runAutomatically == true) // make sure we are still meant to be running automatically
                {
                    updateGrid(); // run a timestep and update visual representation
                    backgroundWorker2.RunWorkerAsync(); // get back to work (sleeping)
                }
            }
        }

        // action taken when the speed trackbar has its value changed - set the new speed
        private void trackBar1_ValueChanged(object sender, EventArgs e)
        {
            speed = trackBar1.Value;
        }

        // action taken when the pause button is clicked
        private void pauseButton_Click(object sender, EventArgs e)
        {
            lock (autolock) // obtain lock - can't suddenly pause else we could leave simulation is unsafe state
            {
                runAutomatically = false; // change state
                // enable buttons
                ManualTimestepButton.Enabled = true;
                trackBar1.Enabled = true;
                runAutoButton.Enabled = true;
                timeStepTextBox.Enabled = true;
                // hide pause
                pauseButton.Visible = false;
            }
        }

        // action taken when the NAR or value button is clicked
        private void NAROrValueButton_Click(object sender, EventArgs e)
        {
            if (showingNAR == false) // check current state
            {
                NAROrValueButton.Text = "Show count value"; // change text appropriately
                showingNAR = !showingNAR; // set state
            }
            else
            {
                NAROrValueButton.Text = "Show NAR value";// change text appropriately
                showingNAR = !showingNAR; // set state
            }
            displayWorldData(); // update the visual representation
        }

        // action when the new simulation menustrip button is clicked - call container form method
        private void startNewSimulationToolStripMenuItem_Click(object sender, EventArgs e)
        {
            pauseButton_Click(pauseButton, null);
            container.showNewWorldScreen();
        }

        /// <summary>
        /// Method for creating a new world. Tells the engine to create a new world and tells the grid to change itself according to the new world's data and then updates the new grid visual display
        /// </summary>
        /// <param name="length"> The new worlds length</param>
        /// <param name="height"> The new worlds height</param>
        /// <param name="bands"> The number of starting bands in the new world</param>
        /// <param name="startingMembers"> The number of starting members in each of the new bands</param>
        public void newWorld(int length, int height, int bands, int startingMembers)
        {
            engine.createNewWorld(length, height, bands, startingMembers);
            resize(length, height);
            grid.newWorldGrid(length, height);
            displayWorldData();
            this.Update();
        }

        /// <summary>
        /// method for extending the size of the form when a large grid is being simulated
        /// </summary>
        /// <param name="length"> length of the grid</param>
        /// <param name="height"> height of the grid</param>
        private void resize(int length, int height)
        {
            int cellSize = 64;
            if (length > 14 || height > 10)
            {
                cellSize = 54;
            }

            if (length > 14)
            {
                this.Size = new Size(380 + length * cellSize, this.Size.Height);
            }
            if (height > 10)
            {
                this.Size = new Size(this.Size.Width, 50 + height * cellSize);
            }
        }

        // action taken when the statistics button is clicked - calls the function in the container form to display the statistics screen
        private void statsButton_Click(object sender, EventArgs e)
        {
            pauseButton_Click(pauseButton, null);
            container.showStatsScreen();
        }

    }
}
